﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using PalmSens.Comm;
using PalmSens.Devices;
using PalmSens.Windows.Devices;

namespace PalmSens.Core.Simplified.WinForms
{
    class DeviceHandler
    {
        internal DeviceHandler()
        {}

        internal Device[] ConnectedDevices { get { return ScanDevices(); } }

        internal bool EnableBluetooth = false;

        internal bool EnableSerialPort = false;

        /// <summary>
        /// Scans for connected devices.
        /// </summary>
        /// <returns>
        /// Returns an array of connected devices
        /// </returns>
        /// <exception cref="System.ArgumentException">An error occured while attempting to scan for connected devices.</exception>
        private Device[] ScanDevices()
        {
            Device[] devices = new Device[0];
            string errors = "";

            try //Attempt to find connected palmsens/emstat devices
            {
                List<DeviceList.DiscoverDevicesFunc> discFuncs = new List<DeviceList.DiscoverDevicesFunc>();
                //Add delegates to list for finding devices on specific communication protocols
                discFuncs.Add(USBCDCDevice.DiscoverDevices); //Default for PS4
                discFuncs.Add(FTDIDevice.DiscoverDevices); //Default for Emstat + PS3
                if(EnableSerialPort)
                    discFuncs.Add(SerialPortDevice.DiscoverDevices); //Devices connected via serial port
                if (EnableBluetooth)
                    discFuncs.Add(BluetoothDevice.DiscoverDevices); //Bluetooth devices (PS4, PS3, Emstat Blue)
                //Return a new array of connected devices found with the included delegate functions 
                devices = new DeviceList(discFuncs).GetAvailableDevices(out errors, EnableBluetooth);
            }
            catch (Exception)
            {
                throw new ArgumentException($"An error occured while attempting to scan for connected devices. {Environment.NewLine} {errors}");
            }
            return devices;
        }

        /// <summary>
        /// Connects to the first device in the connected devices array and returns its CommManager.
        /// </summary>
        /// <returns>
        /// The CommManager of the device it connected to or null
        /// </returns>
        /// <exception cref="System.Exception">Could not find a device to connect to.</exception>
        internal CommManager Connect()
        {
            CommManager comm = null;
            Device[] devices = ScanDevices();
            if (devices.Length == 0)
                throw new Exception("Could not find a device to connect to.");
            else
                comm = Connect(devices[0]);
            return comm;
        }

        /// <summary>
        /// Connects to the specified device and returns its CommManager.
        /// </summary>
        /// <param name="device">The device.</param>
        /// <returns>
        /// The CommManager of the device or null
        /// </returns>
        /// <exception cref="System.ArgumentNullException">The specified device cannot be null.</exception>
        /// <exception cref="System.Exception">Could not connect to the specified device.</exception>
        internal CommManager Connect(Device device)
        {
            if (device == null)
                throw new ArgumentNullException("The specified device cannot be null.");
            CommManager comm = null;

            try
            {
                device.Open(); //Open the device to allow a connection
                comm = new CommManager(device); //Connect to the selected device
            }
            catch (Exception ex)
            {
                device.Close();
                throw new Exception($"Could not connect to the specified device. {ex.Message}");
            }

            return comm;
        }

        /// <summary>
        /// Disconnects the device using its CommManager.
        /// </summary>
        /// <param name="comm">The device's CommManager.</param>
        /// <exception cref="System.ArgumentNullException">The specified CommManager cannot be null.</exception>
        internal void Disconnect(CommManager comm)
        {
            if (comm == null)
                throw new ArgumentNullException("The specified CommManager cannot be null.");
            comm.Disconnect();
            comm = null;
        }
    }
}
